@@ -1,3 +1,7 @@
+# pycargoebuild
+# (c) 2022-2025 Michał Górny <mgorny@gentoo.org>
+# SPDX-License-Identifier: GPL-2.0-or-later
+
 import argparse
 import datetime
 import io
@@ -5,13 +9,12 @@ import json
 import logging
 import lzma
 import os.path
-import shutil
+import stat
 import subprocess
 import sys
 import tarfile
 import tempfile
 import typing
-
 from pathlib import Path, PurePosixPath
 
 if sys.version_info >= (3, 11):
@@ -19,17 +22,25 @@ if sys.version_info >= (3, 11):
 else:
     import tomli as tomllib
 
-from pycargoebuild.cargo import (Crate,
-                                 FileCrate,
-                                 get_crates,
-                                 get_package_metadata,
-                                 )
+from pycargoebuild.cargo import (
+    Crate,
+    FileCrate,
+    WorkspaceCargoTomlError,
+    get_crates,
+    get_package_metadata,
+)
 from pycargoebuild.ebuild import get_ebuild, update_ebuild
-from pycargoebuild.fetch import (fetch_crates_using_wget,
-                                 fetch_crates_using_aria2,
-                                 verify_crates)
-from pycargoebuild.license import load_license_mapping, MAPPING
-
+from pycargoebuild.fetch import (
+    ChecksumMismatchError,
+    fetch_crates_using_aria2,
+    fetch_crates_using_wget,
+    verify_crates,
+)
+from pycargoebuild.license import (
+    MAPPING,
+    UnmatchedLicense,
+    load_license_mapping,
+)
 
 FETCHERS = ("aria2", "wget")
 
@@ -45,6 +56,9 @@ def main(prog_name: str, *argv: str) -> int:
                       action="store_true",
                       help="Pack fetched crates into a tarball rather than "
                            "adding them to the CRATES variable")
+    argp.add_argument("-e", "--features",
+                      action="store_true",
+                      help="Add USE flags for Cargo features")
     argp.add_argument("--crate-tarball-path",
                       default="{distdir}/{name}-{version}-crates.tar.xz",
                       help="Path to write crate tarball to (default: "
@@ -53,6 +67,10 @@ def main(prog_name: str, *argv: str) -> int:
                       default="cargo_home/gentoo",
                       help="Prefix prepended for all paths in the crate "
                            "tarball (default: cargo_home/gentoo)")
+    argp.add_argument("--no-write-crate-tarball",
+                      action="store_true",
+                      help="Do not create the crate tarball, just write "
+                           "the ebuild assuming it exists")
     argp.add_argument("-d", "--distdir",
                       type=Path,
                       help="Directory to store downloaded crates in "
@@ -201,14 +219,16 @@ def main(prog_name: str, *argv: str) -> int:
                       crates: typing.Set[Crate],
                       ) -> None:
         prefix = args.crate_tarball_prefix
+        start_time = datetime.datetime.now(tz=datetime.timezone.utc)
         interval = datetime.timedelta(seconds=10)
-        next_ping = datetime.datetime.now() + interval
+        next_ping = start_time + interval
         for crate_no, crate in enumerate(sorted(crates,
                                                 key=lambda x: x.filename)):
-            if datetime.datetime.now() > next_ping:
+            now = datetime.datetime.now(tz=datetime.timezone.utc)
+            if now > next_ping:
                 logging.info(
                     f"Processed {crate_no} out of {len(crates)} crates")
-                next_ping = datetime.datetime.now() + interval
+                next_ping = now + interval
             if isinstance(crate, FileCrate):
                 with tarfile.open(args.distdir / crate.filename,
                                   "r:gz") as tar_in:
@@ -235,15 +255,33 @@ def main(prog_name: str, *argv: str) -> int:
                     checksum_info.mode = 0o644
                     tar_out.addfile(checksum_info,
                                     io.BytesIO(checksum_data.encode()))
+        end_time = datetime.datetime.now(tz=datetime.timezone.utc)
+        logging.info(f"Time elapsed during repacking: {end_time - start_time}")
 
     crates: typing.Set[Crate] = set()
     pkg_metas = []
     for directory in args.directory:
-        with open(directory / "Cargo.toml", "rb") as f:
+        try:
+            f = open(directory / "Cargo.toml", "rb")
+        except FileNotFoundError:
+            logging.error(f"'Cargo.toml' not found in {str(directory)!r}")
+            logging.info(
+                "Please pass the path to a directory containing 'Cargo.toml' "
+                "as an argument.")
+            return 1
+        with f:
             workspace = get_workspace_root(directory)
             crates.update(workspace.crates)
-            pkg_metas.append(
-                get_package_metadata(f, workspace.workspace_metadata))
+            try:
+                pkg_metas.append(
+                    get_package_metadata(f, workspace.workspace_metadata))
+            except WorkspaceCargoTomlError as e:
+                logging.error("The specified directory is a workspace root: "
+                              f"{str(directory)!r}")
+                logging.info(
+                    "Please run pycargoebuild in one of its members: "
+                    f"{' '.join(e.members)}")
+                return 1
     pkg_meta = pkg_metas[0]
 
     if args.no_license:
@@ -275,7 +313,14 @@ def main(prog_name: str, *argv: str) -> int:
             return 1
 
     fetch_crates(crates)
-    verify_crates(crates, distdir=args.distdir)
+    try:
+        verify_crates(crates, distdir=args.distdir)
+    except ChecksumMismatchError as e:
+        logging.error(f"Checksum mismatch for {str(e.path)!r}")
+        logging.info(f"   Found checksum (SHA256): {e.current!r}")
+        logging.info(f"Expected checksum (SHA256): {e.expected!r}")
+        logging.info("Remove the file to try downloading again.")
+        return 1
 
     umask = os.umask(0)
     os.umask(umask)
@@ -285,57 +330,81 @@ def main(prog_name: str, *argv: str) -> int:
             args.crate_tarball_path.format(name=pkg_meta.name,
                                            version=pkg_meta.version,
                                            distdir=args.distdir))
-        if not args.force and crate_tarball.exists():
-            logging.error(f"{crate_tarball} exists already, pass -f to "
-                          "overwrite it")
-            return 1
-        with tempfile.NamedTemporaryFile(mode="wb",
-                                         dir=args.distdir,
-                                         delete=False
-                                         ) as cratef:
-            try:
-                # typing: https://github.com/python/typeshed/issues/11072
-                with tarfile.open(fileobj=cratef,
-                                  mode="w:xz",
-                                  format=tarfile.GNU_FORMAT,
-                                  encoding="UTF-8",
-                                  preset=9 | lzma.PRESET_EXTREME,
-                                  ) as tar_out:  # type: ignore
-                    os.fchmod(cratef.fileno(), 0o666 & ~umask)
-                    logging.info("Repacking crates ...")
-                    repack_crates(tar_out, crates)
-            except BaseException:
-                Path(cratef.name).unlink()
-                raise
-        Path(cratef.name).rename(crate_tarball)
-        logging.info(f"Crate tarball written to {crate_tarball}")
+        if args.no_write_crate_tarball:
+            logging.info("Skipping creating crate tarball")
+        else:
+            if not args.force and crate_tarball.exists():
+                logging.error(f"{crate_tarball} exists already, pass -f to "
+                              "overwrite it")
+                return 1
+            with tempfile.NamedTemporaryFile(mode="wb",
+                                             dir=args.distdir,
+                                             delete=False
+                                             ) as cratef:
+                try:
+                    # typing: https://github.com/python/typeshed/issues/11072
+                    with tarfile.open(fileobj=cratef,
+                                      mode="w:xz",
+                                      format=tarfile.GNU_FORMAT,
+                                      encoding="UTF-8",
+                                      preset=9 | lzma.PRESET_EXTREME,  # type: ignore
+                                      ) as tar_out:  # type: ignore
+                        os.fchmod(cratef.fileno(), 0o666 & ~umask)
+                        logging.info("Repacking crates ...")
+                        repack_crates(tar_out, crates)
+                except BaseException:
+                    Path(cratef.name).unlink()
+                    raise
+            Path(cratef.name).rename(crate_tarball)
+            logging.info(f"Crate tarball written to {crate_tarball}")
 
-        # do not regenerate Manifest, crate tarball needs to be uploaded
-        # first
-        args.no_manifest = True
+            # do not regenerate Manifest, crate tarball needs to be uploaded
+            # first
+            args.no_manifest = True
 
-    if args.input is not None:
-        ebuild = update_ebuild(
-            args.input.read(),
-            pkg_meta,
-            crates,
-            distdir=args.distdir,
-            crate_license=not args.no_license,
-            crate_tarball=crate_tarball if args.crate_tarball else None,
-            license_overrides=config_toml.get("license-overrides", {}),
-            )
-        logging.warning(
-            "The in-place mode updates CRATES, GIT_CRATES and crate "
-            "LICENSE+= variables only, other metadata is left unchanged")
-    else:
-        ebuild = get_ebuild(
-            pkg_meta,
-            crates,
-            distdir=args.distdir,
-            crate_license=not args.no_license,
-            crate_tarball=crate_tarball if args.crate_tarball else None,
-            license_overrides=config_toml.get("license-overrides", {}),
-            )
+    try:
+        if args.input is not None:
+            ebuild = update_ebuild(
+                args.input.read(),
+                pkg_meta,
+                crates,
+                distdir=args.distdir,
+                crate_license=not args.no_license,
+                crate_tarball=crate_tarball if args.crate_tarball else None,
+                license_overrides=config_toml.get("license-overrides", {}),
+                )
+            logging.warning(
+                "The in-place mode updates CRATES, GIT_CRATES and crate "
+                "LICENSE+= variables only, other metadata is left unchanged")
+        else:
+            ebuild = get_ebuild(
+                pkg_meta,
+                crates,
+                distdir=args.distdir,
+                crate_license=not args.no_license,
+                crate_tarball=crate_tarball if args.crate_tarball else None,
+                license_overrides=config_toml.get("license-overrides", {}),
+                use_features=args.features,
+                )
+    except UnmatchedLicense as e:
+        logging.error(
+            f"The license {e.license_key!r} did not match any entry in "
+            f"{args.license_mapping.name!r}")
+        if e.crate is not None:
+            logging.info(f"Crate file: {e.crate!r}")
+        logging.info(
+            "1. If that is a valid SPDX-2.0 license identifier, then please "
+            "add it to the license mapping file.  However, please make sure "
+            "to:\n"
+            "a. avoid adding duplicate licenses (multiple SPDX-2.0 "
+            "identifiers can map to the same Gentoo license),\n"
+            "b. add new licenses to appropriate license-groups.\n"
+            "\n"
+            "2. If that is not a valid SPDX-2.0 license identiiers, please "
+            "file a bug upstream.  For the time being, you can use a local "
+            "license mapping file (--license-mapping) or per-crate "
+            "license-overrides in config (see README).")
+        return 1
 
     with tempfile.NamedTemporaryFile(mode="w",
                                      encoding="utf-8",
@@ -343,10 +412,9 @@ def main(prog_name: str, *argv: str) -> int:
                                      delete=False) as outf:
         try:
             if args.input is not None:
-                # typeshed is missing fd support in shutil.copymode()
-                # https://github.com/python/typeshed/issues/9288
-                shutil.copymode(args.input.fileno(),
-                                outf.fileno())  # type: ignore
+                st = os.stat(args.input.fileno())
+                os.chown(outf.fileno(), st.st_uid, st.st_gid)
+                os.chmod(outf.fileno(), stat.S_IMODE(st.st_mode))
                 args.input.close()
             else:
                 os.fchmod(outf.fileno(), 0o666 & ~umask)
